from collections import Counter, defaultdict
from functools import reduce
from itertools import chain
from operator import mul
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Type, TypeVar, cast

from sympy import Expr, Function, var

from comb_spec_searcher import (
    AtomStrategy,
    CombinatorialClass,
    StrategyPack,
    VerificationStrategy,
)
from comb_spec_searcher.exception import InvalidOperationError, StrategyDoesNotApply
from comb_spec_searcher.typing import Objects, Terms
from permuta import Perm
from permuta.permutils import (
    is_insertion_encodable_maximum,
    is_insertion_encodable_rightmost,
)
from permuta.permutils.symmetry import all_symmetry_sets
from tilings import GriddedPerm, Tiling
from tilings.algorithms import locally_factorable_shift
from tilings.algorithms.enumeration import (
    DatabaseEnumeration,
    LocalEnumeration,
    MonotoneTreeEnumeration,
)
from tilings.assumptions import ComponentAssumption
from tilings.strategies import (
    FactorFactory,
    FactorInsertionFactory,
    RequirementCorroborationFactory,
)

x = var("x")

__all__ = [
    "BasisAwareVerificationStrategy",
    "BasicVerificationStrategy",
    "OneByOneVerificationStrategy",
    "DatabaseVerificationStrategy",
    "LocallyFactorableVerificationStrategy",
    "ElementaryVerificationStrategy",
    "LocalVerificationStrategy",
    "MonotoneTreeVerificationStrategy",
]


TileScopeVerificationStrategy = VerificationStrategy[Tiling, GriddedPerm]
BasisAwareVerificationStrategyType = TypeVar(
    "BasisAwareVerificationStrategyType", bound="BasisAwareVerificationStrategy"
)


class BasisAwareVerificationStrategy(TileScopeVerificationStrategy):
    """
    A base class for a verification strategy that needs to know the basis the
    Tilescope is currently running.
    """

    def __init__(
        self,
        basis: Optional[Iterable[Perm]] = None,
        symmetry: bool = False,
        ignore_parent: bool = True,
    ):
        self._basis = tuple(basis) if basis is not None else tuple()
        self._symmetry = symmetry
        assert all(
            isinstance(p, Perm) for p in self._basis
        ), "Element of the basis must be Perm"
        if symmetry:
            self.symmetries = frozenset(
                frozenset(b) for b in all_symmetry_sets(self._basis)
            )
        else:
            self.symmetries = frozenset([frozenset(self._basis)])
        super().__init__(ignore_parent=ignore_parent)

    def change_basis(
        self: BasisAwareVerificationStrategyType, basis: Iterable[Perm], symmetry: bool
    ) -> BasisAwareVerificationStrategyType:
        """
        Return a new version of the verification strategy with the given basis instead
        of the current one.
        """
        basis = tuple(basis)
        return self.__class__(basis, symmetry, self.ignore_parent)

    @property
    def basis(self) -> Tuple[Perm, ...]:
        return self._basis

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["basis"] = self._basis
        d["symmetry"] = self._symmetry
        return d

    @classmethod
    def from_dict(
        cls: Type[BasisAwareVerificationStrategyType], d: dict
    ) -> BasisAwareVerificationStrategyType:
        if "basis" in d and d["basis"] is not None:
            basis: Optional[List[Perm]] = [Perm(p) for p in d.pop("basis")]
        else:
            basis = d.pop("basis", None)
        return cls(basis=basis, **d)

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"basis={self._basis}",
                f"symmetry={self._symmetry}",
                f"ignore_parent={self.ignore_parent}",
            ]
        )
        return f"{self.__class__.__name__}({args})"


class BasicVerificationStrategy(AtomStrategy):
    """
    TODO: can this be moved to the CSS atom strategy?
    """

    @staticmethod
    def get_terms(comb_class: CombinatorialClass, n: int) -> Terms:
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        gp = next(comb_class.minimal_gridded_perms())
        if n == len(gp):
            parameters = tuple(
                assumption.get_value(gp) for assumption in comb_class.assumptions
            )
            return Counter([parameters])
        return Counter()

    @staticmethod
    def get_objects(comb_class: CombinatorialClass, n: int) -> Objects:
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        res: Objects = defaultdict(list)
        gp = next(comb_class.minimal_gridded_perms())
        if n == len(gp):
            parameters = tuple(
                assumption.get_value(gp) for assumption in comb_class.assumptions
            )
            res[parameters].append(gp)
        return res

    @staticmethod
    def generate_objects_of_size(
        comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        """
        Verification strategies must contain a method to generate the objects.
        """
        yield from comb_class.objects_of_size(n, **parameters)

    @staticmethod
    def random_sample_object_of_size(
        comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> GriddedPerm:
        """
        Verification strategies must contain a method to sample the objects.
        """
        key = tuple(y for _, y in sorted(parameters.items()))
        if BasicVerificationStrategy.get_terms(comb_class, n).get(key):
            return cast(GriddedPerm, next(comb_class.objects_of_size(n, **parameters)))

    def get_genf(
        self,
        comb_class: CombinatorialClass,
        funcs: Optional[Dict[CombinatorialClass, Function]] = None,
    ) -> Expr:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("Can't find generating functon for non-atom.")
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        cast(Tiling, comb_class)
        gp = next(comb_class.minimal_gridded_perms())
        expected = {"x": len(gp)}
        for assumption in comb_class.assumptions:
            expected[
                comb_class.get_assumption_parameter(assumption)
            ] = assumption.get_value(gp)
        return reduce(mul, [var(k) ** n for k, n in expected.items()], 1)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class OneByOneVerificationStrategy(BasisAwareVerificationStrategy):
    @staticmethod
    def pack(comb_class: Tiling) -> StrategyPack:
        if any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions):
            raise InvalidOperationError(
                "Can't find generating function with component assumption."
            )
        # pylint: disable=import-outside-toplevel
        from tilings.tilescope import TileScopePack

        assert comb_class.dimensions == (1, 1)
        basis, _ = comb_class.cell_basis()[(0, 0)]
        if any(
            any(p.contains(patt) for patt in basis)
            for p in [
                Perm((0, 2, 1)),
                Perm((1, 2, 0)),
                Perm((1, 0, 2)),
                Perm((2, 0, 1)),
            ]
        ):
            # subclass of Av(231) or a symmetry, use point placements!
            return TileScopePack.point_and_row_and_col_placements().add_verification(
                BasicVerificationStrategy(), replace=True
            )
        if is_insertion_encodable_maximum(basis):
            return TileScopePack.regular_insertion_encoding(3)
        if is_insertion_encodable_rightmost(basis):
            return TileScopePack.regular_insertion_encoding(2)
        # if it is the class or positive class
        if not comb_class.requirements or (
            len(comb_class.requirements) == 1
            and len(comb_class.requirements[0]) == 1
            and len(comb_class.requirements[0][0]) <= 2
        ):
            if basis in ([Perm((0, 1, 2))], [Perm((2, 1, 0))]):
                # Av(123) or Av(321) - use fusion!
                return (
                    TileScopePack.row_and_col_placements(row_only=True)
                    .make_fusion(tracked=True)
                    .add_basis(basis)
                )
            if (Perm((0, 1, 2)) in basis or Perm((2, 1, 0)) in basis) and all(
                len(p) <= 4 for p in basis
            ):
                # is a subclass of Av(123) avoiding patterns of length <= 4
                # experimentally showed that such clsses always terminates
                return TileScopePack.row_and_col_placements().add_basis(basis)
        raise InvalidOperationError(
            "Cannot get a specification for one by one verification for "
            f"subclass Av({basis})"
        )

    def verified(self, comb_class: Tiling) -> bool:
        return comb_class.dimensions == (1, 1) and (
            frozenset(ob.patt for ob in comb_class.obstructions) not in self.symmetries
            or any(
                isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions
            )
        )

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Expr:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not 1x1 verified")
        if len(comb_class.obstructions) == 1 and comb_class.obstructions[0] in (
            GriddedPerm.single_cell((0, 1, 2), (0, 0)),
            GriddedPerm.single_cell((2, 1, 0), (0, 0)),
        ):
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)

    @staticmethod
    def formal_step() -> str:
        return "tiling is a subclass of the original tiling"

    @staticmethod
    def get_terms(comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for one by one verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for one by one "
            "verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for one by one verified tilings"
        )

    def __str__(self) -> str:
        return "one by one verification"


class DatabaseVerificationStrategy(TileScopeVerificationStrategy):
    """
    Enumeration strategy for a tilings that are in the database.

    There is not always a specification for a tiling in the database but you
    can always find the generating function by looking up the database.
    """

    @staticmethod
    def pack(comb_class: Tiling) -> StrategyPack:
        # TODO: check database for tiling
        raise InvalidOperationError(
            "Cannot get a specification for a tiling in the database"
        )

    @staticmethod
    def verified(comb_class: Tiling):
        return DatabaseEnumeration(comb_class).verified()

    @staticmethod
    def formal_step() -> str:
        return "tiling is in the database"

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Expr:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling is not in the database")
        return DatabaseEnumeration(comb_class).get_genf()

    @staticmethod
    def get_terms(comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for database verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for database verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for database verified tilings"
        )

    def __str__(self) -> str:
        return "database verification"

    @classmethod
    def from_dict(cls, d: dict) -> "DatabaseVerificationStrategy":
        return cls(**d)


class LocallyFactorableVerificationStrategy(BasisAwareVerificationStrategy):
    """
    Verification strategy for a locally factorable tiling.

    A tiling is locally factorable if all its obstructions and requirements are
    locally factorable, i.e. each obstruction or requirement use at most one
    cell on each row and column. To be locally factorable, a tiling
    should not be equivalent to a 1x1 tiling.

    A locally factorable tiling can be describe with a specification with only subset
    verified tiling.
    """

    @staticmethod
    def pack(comb_class: Tiling) -> StrategyPack:
        if any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions):
            raise InvalidOperationError(
                "Can't find generating function with component assumption."
            )
        return StrategyPack(
            name="LocallyFactorable",
            initial_strats=[FactorFactory(), RequirementCorroborationFactory()],
            inferral_strats=[],
            expansion_strats=[[FactorInsertionFactory()]],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
                LocalVerificationStrategy(no_factors=True),
            ],
        )

    @staticmethod
    def _locally_factorable_obstructions(tiling: Tiling):
        """
        Check if all the obstructions of the tiling are locally factorable.
        """
        return all(not ob.is_interleaving() for ob in tiling.obstructions)

    @staticmethod
    def _locally_factorable_requirements(tiling: Tiling):
        """
        Check if all the requirements of the tiling are locally factorable.
        """
        reqs = chain.from_iterable(tiling.requirements)
        return all(not r.is_interleaving() for r in reqs)

    def verified(self, comb_class: Tiling):
        return (
            not comb_class.dimensions == (1, 1)
            and self._locally_factorable_obstructions(comb_class)
            and self._locally_factorable_requirements(comb_class)
        )

    def decomposition_function(
        self, comb_class: Tiling
    ) -> Optional[Tuple[Tiling, ...]]:
        """
        The rule as the root as children if one of the cell of the tiling is the root.
        """
        if self.verified(comb_class):
            if not self.basis:
                return ()
            sfs = locally_factorable_shift.shift_from_spec(
                comb_class, self, self.symmetries
            )
            if sfs is not None:
                return (Tiling.from_perms(self.basis),)
            return ()
        return None

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply
        if not children:
            return ()
        shift = locally_factorable_shift.shift_from_spec(
            comb_class, self, self.symmetries
        )
        assert shift is not None
        return (shift,)

    @staticmethod
    def formal_step() -> str:
        return "tiling is locally factorable"

    def __str__(self) -> str:
        return "locally factorable verification"


class ElementaryVerificationStrategy(LocallyFactorableVerificationStrategy):
    """
    Verification strategy for elementary tilings.

    A tiling is elementary if each active cell is on its own row and column.
    To be elementary, a tiling should not be equivalent to a 1x1
    tiling.

    By definition an elementary tiling is locally factorable.

    A elementary tiling can be describe with a specification with only one by one
    verified tiling.
    """

    @staticmethod
    def verified(comb_class: Tiling):
        return comb_class.fully_isolated() and not comb_class.dimensions == (1, 1)

    @staticmethod
    def formal_step() -> str:
        return "tiling is elementary verified"

    @classmethod
    def from_dict(cls, d: dict) -> "ElementaryVerificationStrategy":
        return cls(**d)

    def __str__(self) -> str:
        return "elementary verification"


class LocalVerificationStrategy(TileScopeVerificationStrategy):
    """
    The local verified strategy.

    A tiling is local verified if every obstruction and every requirement is
    localized, i.e. in a single cell and the tiling is not 1x1.
    """

    def __init__(self, ignore_parent: bool = True, no_factors: bool = False):
        self.no_factors = no_factors
        super().__init__(ignore_parent=ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        try:
            return InsertionEncodingVerificationStrategy().pack(comb_class)
        except StrategyDoesNotApply:
            pass
        if self.no_factors:
            raise InvalidOperationError("Cannot get a simpler specification")
        if (
            any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions)
            and len(comb_class.find_factors()) == 1
        ):
            raise InvalidOperationError(
                "Can't find generating function with component assumption."
            )
        return StrategyPack(
            initial_strats=[FactorFactory()],
            inferral_strats=[],
            expansion_strats=[],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
                LocalVerificationStrategy(no_factors=True),
            ],
            name="factor pack",
        )

    def verified(self, comb_class: Tiling) -> bool:
        return (
            comb_class.dimensions != (1, 1)
            and (not self.no_factors or len(comb_class.find_factors()) == 1)
            and LocalEnumeration(comb_class).verified()
        )

    @staticmethod
    def formal_step() -> str:
        return "tiling is locally enumerable"

    @classmethod
    def from_dict(cls, d: dict) -> "LocalVerificationStrategy":
        return cls(**d)

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Expr:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not locally verified")
        if len(comb_class.obstructions) == 1 and comb_class.obstructions[0] in (
            GriddedPerm.single_cell((0, 1, 2), (0, 0)),
            GriddedPerm.single_cell((2, 1, 0), (0, 0)),
        ):
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)

    @staticmethod
    def get_terms(comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for locally verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for locally verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for locally verified tilings"
        )

    def __str__(self) -> str:
        return "local verification"


class InsertionEncodingVerificationStrategy(TileScopeVerificationStrategy):
    """
    Verify all n x 1 and 1 x n tilings that have a regular insertion encoding.
    """

    def __init__(self, ignore_parent: bool = True):
        super().__init__(ignore_parent=ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        if any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions):
            raise InvalidOperationError(
                "Can't find generating function with component assumption."
            )
        # pylint: disable=import-outside-toplevel
        from tilings.strategy_pack import TileScopePack

        if self.has_rightmost_insertion_encoding(comb_class):
            return TileScopePack.regular_insertion_encoding(2)
        if self.has_topmost_insertion_encoding(comb_class):
            return TileScopePack.regular_insertion_encoding(3)
        raise StrategyDoesNotApply("tiling does not has a regular insertion encoding")

    @staticmethod
    def has_rightmost_insertion_encoding(tiling: Tiling) -> bool:
        return tiling.dimensions[0] == 1 and all(
            is_insertion_encodable_rightmost(basis)
            for basis, _ in tiling.cell_basis().values()
        )

    @staticmethod
    def has_topmost_insertion_encoding(tiling: Tiling) -> bool:
        return tiling.dimensions[1] == 1 and all(
            is_insertion_encodable_maximum(basis)
            for basis, _ in tiling.cell_basis().values()
        )

    def verified(self, comb_class: Tiling) -> bool:
        return self.has_rightmost_insertion_encoding(
            comb_class
        ) or self.has_topmost_insertion_encoding(comb_class)

    @staticmethod
    def formal_step() -> str:
        return "tiling has a regular insertion encoding"

    @classmethod
    def from_dict(cls, d: dict) -> "InsertionEncodingVerificationStrategy":
        return cls(**d)

    @staticmethod
    def get_terms(comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for insertion encoding "
            "verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for insertion encoding "
            "verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for insertion encoding verified tilings"
        )

    def __str__(self) -> str:
        return "insertion encoding verified"


class MonotoneTreeVerificationStrategy(TileScopeVerificationStrategy):
    """
    Verify all tiling that is a monotone tree.
    """

    def __init__(self, ignore_parent: bool = True, no_factors: bool = True):
        self.no_factors = no_factors
        super().__init__(ignore_parent=ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        if any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions):
            raise InvalidOperationError(
                "Can't find generating function with component assumption."
            )
        try:
            return InsertionEncodingVerificationStrategy().pack(comb_class)
        except StrategyDoesNotApply:
            pass
        if self.no_factors:
            raise InvalidOperationError("Cannot get a simpler specification")
        return StrategyPack(
            initial_strats=[FactorFactory()],
            inferral_strats=[],
            expansion_strats=[],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
            ],
            name="factor pack",
        )

    def verified(self, comb_class: Tiling) -> bool:
        return (
            not self.no_factors or len(comb_class.find_factors()) == 1
        ) and MonotoneTreeEnumeration(comb_class).verified()

    @staticmethod
    def formal_step() -> str:
        return "tiling is a monotone tree"

    @classmethod
    def from_dict(cls, d: dict) -> "MonotoneTreeVerificationStrategy":
        return cls(**d)

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Expr:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not locally verified")
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return MonotoneTreeEnumeration(comb_class).get_genf(funcs=funcs)

    @staticmethod
    def get_terms(comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for monotone tree "
            "verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for monotone tree "
            "verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for monotone tree verified tilings"
        )

    def __str__(self) -> str:
        return "monotone tree verification"



from typing import Iterator, Optional, Tuple

from comb_spec_searcher import StrategyFactory
from comb_spec_searcher.strategies import Rule
from tilings import GriddedPerm, Tiling
from tilings.algorithms import ComponentFusion, Fusion

from .fusion import FusionStrategy


class ComponentFusionStrategy(FusionStrategy):
    def fusion_algorithm(self, tiling: Tiling) -> Fusion:
        return ComponentFusion(
            tiling, row_idx=self.row_idx, col_idx=self.col_idx, tracked=self.tracked
        )

    def formal_step(self) -> str:
        fusing = "rows" if self.row_idx is not None else "columns"
        idx = self.row_idx if self.row_idx is not None else self.col_idx
        return "component fuse {} {} and {}".format(fusing, idx, idx + 1)

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
        left_points: Optional[int] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        raise NotImplementedError


class ComponentFusionFactory(StrategyFactory[Tiling]):
    def __init__(self, tracked: bool = False, isolation_level: Optional[str] = None):
        self.tracked = tracked
        self.isolation_level = isolation_level

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        if comb_class.requirements:
            return
        cols, rows = comb_class.dimensions
        for row_idx in range(rows - 1):
            algo = ComponentFusion(
                comb_class,
                row_idx=row_idx,
                tracked=self.tracked,
                isolation_level=self.isolation_level,
            )
            if algo.fusable():
                fused_tiling = algo.fused_tiling()
                yield ComponentFusionStrategy(row_idx=row_idx, tracked=self.tracked)(
                    comb_class, (fused_tiling,)
                )
        for col_idx in range(cols - 1):
            algo = ComponentFusion(
                comb_class,
                col_idx=col_idx,
                tracked=self.tracked,
                isolation_level=self.isolation_level,
            )
            if algo.fusable():
                fused_tiling = algo.fused_tiling()
                yield ComponentFusionStrategy(col_idx=col_idx, tracked=self.tracked)(
                    comb_class, (fused_tiling,)
                )

    def __str__(self) -> str:
        return f"{'tracked ' if self.tracked else ''}component fusion"

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"(tracked={self.tracked}, isolation_level={self.isolation_level})"
        )

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["tracked"] = self.tracked
        d["isolation_level"] = self.isolation_level
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ComponentFusionFactory":
        return cls(**d)



from collections import Counter, defaultdict
from heapq import heapify, heappop, heappush
from itertools import chain, product
from typing import TYPE_CHECKING, Dict, FrozenSet, Iterator, List, Optional, Set, Tuple

from permuta import Perm
from tilings import GriddedPerm

if TYPE_CHECKING:
    from tilings import Tiling


__all__ = ["MinimalGriddedPerms"]

Cell = Tuple[int, int]
GPTuple = Tuple[GriddedPerm, ...]
Reqs = Tuple[GPTuple, ...]
WorkPackets = Tuple[GriddedPerm, GPTuple, Cell]


class QueuePacket:
    def __init__(
        self,
        gp: GriddedPerm,
        gps: GPTuple,
        last_cell: Cell,
        still_localising: bool,
        mindices: Dict[Cell, int],
    ):
        self.gp = gp
        self.gps = gps
        self.last_cell = last_cell
        self.still_localising = still_localising
        self.mindices = mindices

    def __lt__(self, other: "QueuePacket"):
        return len(self.gp) < len(other.gp)


class MinimalGriddedPerms:
    def __init__(self, tiling: "Tiling"):
        self.obstructions = tiling.obstructions
        self.requirements = tiling.requirements
        self.relevant_obstructions: Dict[FrozenSet[Cell], GPTuple] = dict()
        self.relevant_requirements: Dict[FrozenSet[Cell], Reqs] = dict()
        self.relevant_obstructions_by_cell: Dict[
            Tuple[Cell, FrozenSet[Cell]], GPTuple
        ] = dict()
        self.requirements_up_to_cell: Dict[Tuple[Cell, GPTuple], GPTuple] = dict()
        # Delay computing until needed - these could all be stored on
        # MinimalGriddedPerms, as none of these are tiling specific.
        self.localised_patts: Dict[Tuple[Cell, GPTuple], GPTuple] = dict()
        self.max_cell_counts: Dict[GPTuple, Dict[Cell, int]] = dict()
        self.upward_closures: Dict[GPTuple, GPTuple] = dict()
        self.known_patts: Dict[GriddedPerm, Set[GriddedPerm]] = defaultdict(set)

    def get_requirements_up_to_cell(self, cell: Cell, gps: GPTuple) -> GPTuple:
        """Given a goal gps and cell (x,y), return the truncations of the reqs
        in gps to the cells < (x, y) in normal sort order."""
        res = self.requirements_up_to_cell.get((cell, gps))
        if res is None:
            res = tuple(
                req.get_gridded_perm_in_cells(frozenset(c for c in req.pos if c < cell))
                for req in gps
            )
            self.requirements_up_to_cell[(cell, gps)] = res
        return res

    def get_relevant_obstructions(self, gp: GriddedPerm) -> GPTuple:
        """Get the obstructions that involve only cells in the gp."""
        cells = frozenset(gp.pos)
        res = self.relevant_obstructions.get(cells)
        if res is None:
            res = tuple(
                ob for ob in self.obstructions if all(c in cells for c in ob.pos)
            )
            self.relevant_obstructions[cells] = res
        return res

    def get_relevant_obstructions_by_cell(self, gp: GriddedPerm, cell: Cell) -> GPTuple:
        """Get the obstructions that involve only cells in gp
        and involve cell"""
        cells = frozenset(gp.pos)
        res = self.relevant_obstructions_by_cell.get((cell, cells))
        if res is None:
            relevant_obstructions = self.get_relevant_obstructions(gp)
            res = tuple(ob for ob in relevant_obstructions if cell in ob.pos)
            self.relevant_obstructions_by_cell[(cell, cells)] = res
        return res

    def get_relevant_requirements(self, gp: GriddedPerm) -> Reqs:
        """Get the requirements that involve only cells in gp."""
        cells = frozenset(gp.pos)
        res = self.relevant_requirements.get(cells)
        if res is None:
            res = tuple(
                tuple(req for req in reqlist if all(c in cells for c in req.pos))
                for reqlist in self.requirements
            )
            self.relevant_requirements[cells] = res
        return res

    def satisfies_obstructions(
        self, gp: GriddedPerm, must_contain: Optional[Cell] = None
    ) -> bool:
        """Check if a gridded permutation avoids the obstructions."""
        if must_contain is None:
            obs = self.get_relevant_obstructions(gp)
        else:
            obs = self.get_relevant_obstructions_by_cell(gp, must_contain)
        return gp.avoids(*obs)

    def satisfies_requirements(self, gp: GriddedPerm) -> bool:
        """Check if a gridded permutation contains all requirements."""
        reqs = self.get_relevant_requirements(gp)
        return all(reqs) and all(self.contains(gp, *req) for req in reqs)

    def avoids(self, gp: GriddedPerm, *patts: GriddedPerm) -> bool:
        """An avoidance check. See contains method."""
        return not self.contains(gp, *patts)

    def contains(self, gp: GriddedPerm, *patts: GriddedPerm) -> bool:
        """This is a an alternative containment check.
        It will return True if any element of patts is contained in gp, and add
        it to the known_patts. Only use this if it is worth caching, so in the
        usage case, if the check is related to the target gps requirement."""
        known_patts = self.known_patts[gp]
        if any(patt in known_patts for patt in patts):
            return True
        for patt in patts:
            if gp.contains(patt):
                known_patts.add(patt)
                return True
        return False

    def _prepare_queue(self, queue: List[QueuePacket]) -> Iterator[GriddedPerm]:
        """Add cell counters with gridded permutations to the queue.
        The function yields all initial_gp that satisfy the requirements."""
        if len(self.requirements) <= 1:
            return
        for gps in product(*self.requirements):
            # try to stitch together as much of the independent cells of the
            # gridded permutation together first
            initial_gp = self.initial_gp(*gps)
            if self.satisfies_obstructions(initial_gp):
                # We will now prepare to add it to the queue.
                if self.satisfies_requirements(initial_gp):
                    yield initial_gp
                    continue
                # we pass on this information, together with the target gps
                # will be used to guide us in choosing smartly which cells to
                # insert into - see the 'get_cells_to_try' method.
                # mindices ensure we insert left to right in each cell
                mindices: Dict[Cell, int] = dict()
                # last_cell ensures that we insert into cells in order after
                # we have satisfied the local patterns needed
                last_cell = (-1, -1)
                # this flag ensures we first make sure we contain all local
                # patterns
                localised = True
                # add all of this information the queue. The Info class ensures
                # we sort by initial_gp only.
                new_info = QueuePacket(
                    initial_gp, tuple(gps), last_cell, localised, mindices
                )
                heappush(queue, new_info)

    def initial_gp(self, *gps: GriddedPerm) -> GriddedPerm:
        """
        Return the gridded perm that initialises with the first argument,
        then adds as many points as possible from the left to right, such that
        the cells don't share a row or column. In this instance, to contain
        both you must contain the respective subgridded perm in the independent
        cell.
        """
        # We work with the upward closure.
        gps = self.get_upward_closure(gps)
        # Initialise with the first gridded permutation in gps.
        res = GriddedPerm(gps[0].patt, gps[0].pos)
        active_cols = set(i for i, j in res.pos)
        active_rows = set(j for i, j in res.pos)
        for gp in gps[1:]:
            # for each subsequent gridded perm in gps, find all the cells
            # not on the same row or column
            indep_cells = [
                (i, j)
                for i, j in gp.pos
                if i not in active_cols and j not in active_rows
            ]
            if indep_cells:
                # take the subgp using the independent cells.
                # we will now glue together the result, and the new independent
                # cell in the unique way that this can be done
                subgp = gp.get_gridded_perm_in_cells(indep_cells)
                # update future active rows and cols
                active_cols.update(i for i, j in subgp.pos)
                active_rows.update(j for i, j in subgp.pos)
                # temp will be sorted, and standardised to create the unique
                # minimal gridded permutation containing both res and subgp.
                # We want cells at lower index front - if they are in the same
                # column, then they come from the same gridded permutation, so
                # we sort second by the index of its original gridded perm.
                # We will standardise based on values. Those in lower cells
                # will have smaller value, and if they are in the same row then
                # they come from the same gridded permutation so the second
                # entry is the value from its original gridded perm.
                temp = [
                    ((cell[0], idx), (cell[1], val))
                    for (idx, val), cell in zip(enumerate(res.patt), res.pos)
                ] + [
                    ((cell[0], idx), (cell[1], val))
                    for (idx, val), cell in zip(enumerate(subgp.patt), subgp.pos)
                ]
                temp.sort()
                # update the res
                new_pos = [(idx[0], val[0]) for idx, val in temp]
                new_patt = Perm.to_standard(val for _, val in temp)
                res = GriddedPerm(new_patt, new_pos)
        return res

    @staticmethod
    def cell_counter(*gps: GriddedPerm) -> Dict[Cell, int]:
        """Returns a counter of the sum of number of cells used by the gridded
        permutations."""
        return Counter(chain.from_iterable(gp.pos for gp in gps))

    def get_max_cell_count(self, gps: GPTuple) -> Dict[Cell, int]:
        """Return the theoretical maximum number of points needed in each cell.
        It assumes that gps is closed upwards."""
        res = self.max_cell_counts.get(gps)
        if res is None:
            # we work with the upward closure.
            gps = self.get_upward_closure(gps)
            res = MinimalGriddedPerms.cell_counter(*gps)
            # now we look for any cells containing more than one full req and
            # are not involved with any other requirements, because we may be
            # able to apply a better upper bound to the number of points
            # required in the cell
            for cell in res.keys():
                in_this_cell: Set[Perm] = set()
                for req in gps:
                    # we ignore point requirements
                    if len(req) > 1 and cell in req.pos:
                        if req.is_single_cell():
                            # [req] is entirely in [cell]
                            in_this_cell.add(req.patt)
                        else:
                            # [req] involves [cell], but also some other cell,
                            # so this cell won't have a usable upper bound
                            break
                else:
                    # if we get here, then we can use the precomputed bounds
                    # for this cell
                    better_bounds = MinimalGriddedPerms.better_bounds()
                    # the bounds can only be better if there are two or more
                    # patterns
                    if len(in_this_cell) >= 2:
                        # the dictionary is patt: v, where v is the number we
                        # can subtract from the naive bound
                        res[cell] -= better_bounds.get(frozenset(in_this_cell), 0)
            self.max_cell_counts[gps] = res
        return res

    _better_bounds = None

    @staticmethod
    def better_bounds() -> Dict[FrozenSet[Perm], int]:
        """
        Return a dictionary. The entry k:v means that if a cell contains
        the full requirments in k, and no others touch it, then the upper
        bound for the number of points that will need to be placed in that
        cell is the sum of the lenths of the perms in k, minus v.

        Example: a cell containing 01 and 10 has an upper bound of 3, not 4
        """
        if MinimalGriddedPerms._better_bounds is None:
            # pylint: disable=import-outside-toplevel
            from .better_bounds import better_bounds

            MinimalGriddedPerms._better_bounds = better_bounds
        return MinimalGriddedPerms._better_bounds

    def get_localised_patts(self, cell: Cell, gps: GPTuple) -> GPTuple:
        """Return the localised patts that gps must contain."""
        res = self.localised_patts.get((cell, gps))
        if res is None:
            local_patts = set()
            for gp in gps:
                local_patts.add(gp.get_gridded_perm_in_cells([cell]))
            # Only keep the upward closure.
            res = self.get_upward_closure(tuple(local_patts))
            self.localised_patts[(cell, gps)] = res
        return res

    def get_upward_closure(self, gps: GPTuple) -> GPTuple:
        """Return list of gridded perms such that every gridded perm contained
        in another is removed."""
        res = self.upward_closures.get(gps)
        if res is None:
            upward_closure: List[GriddedPerm] = []
            for gp in sorted(gps, key=len, reverse=True):
                if all(gp not in g for g in upward_closure):
                    upward_closure.append(gp)
            res = tuple(upward_closure)
            self.upward_closures[gps] = res
        return res

    def _get_cells_to_try(self, qpacket: QueuePacket) -> Iterator[Tuple[Cell, bool]]:
        """Yield cells that a gridded permutation could be extended by."""
        cells: Set[Cell] = set()
        for g, req in zip(qpacket.gps, self.get_relevant_requirements(qpacket.gp)):
            if not self.contains(qpacket.gp, *req):
                # Only insert into cells in the requirements that are not
                # already satisfied.
                cells.update(g.pos)
            elif g not in req or self.avoids(qpacket.gp, g):
                # If any requirement contains a requirement, but not the target
                # gridded perm from that requirement in gps, then inserting
                # further will not result in a minimal gridded perm that will
                # not be found by another target.
                return
        # Ensure we insert into the smallest cell possible first.
        currcellcounts = self.cell_counter(qpacket.gp)
        max_cell_count = self.get_max_cell_count(qpacket.gps)

        def can_insert(cell: Cell) -> bool:
            """only consider cells which don't already have the maximum number
            of points inserted into them"""
            return max_cell_count[cell] > currcellcounts[cell]

        possible_cells = sorted(filter(can_insert, cells))

        def try_yield(
            cells: List[Cell], localised: bool
        ) -> Iterator[Tuple[Cell, bool]]:
            for cell in cells:
                # if this comes from a localised pattern, then yield
                # otherwise we can only insert into a cell greater or equal to
                # the last cell inserted into, to avoid duplicate work.
                if localised or cell == qpacket.last_cell:
                    yield (cell, localised)
                # if it is greater, then we must ensure that we can satisfy the
                # the requirements we are aiming for in gps, and contain all of
                # the gps restricted to smaller cells
                elif cell > qpacket.last_cell:
                    prior_reqs = self.get_requirements_up_to_cell(cell, qpacket.gps)
                    if all(self.contains(qpacket.gp, req) for req in prior_reqs):
                        yield (cell, localised)

        if qpacket.still_localising:
            for cell in possible_cells:
                # try to insert into a cell that doesn't contain all of the
                # localised patterns frist
                if self.avoids(
                    qpacket.gp, *self.get_localised_patts(cell, qpacket.gps)
                ):
                    yield from try_yield([cell], True)
                    return
        # otherwise, we are none the wiser which cell to insert into so try
        # all possible cells
        yield from try_yield(possible_cells, False)

    @staticmethod
    def insert_point(
        gp: GriddedPerm, cell: Cell, minimumdex: int
    ) -> Iterator[Tuple[int, GriddedPerm]]:
        """Yield all possible gridded perms where a point is inserted into
        the given cell, where the minimum index is mindex if given. We insert
        right to left, to ensure that if a gp is made twice, that with the
        largest index is yielded first.

        Each gridded perm is yielded as a tuple, consisting of the idx of the
        newly inserted point and the gp."""
        # Find the bounding box of where the point can be inserted.
        mindex, maxdex, minval, maxval = gp.get_bounding_box(cell)
        # We make sure we only insert to the right of where the last
        # point was inserted into the cell.
        mindex = max(mindex, minimumdex)
        for idx, val in product(
            range(maxdex, mindex - 1, -1), range(minval, maxval + 1)
        ):
            nextgp = gp.insert_specific_point(cell, idx, val)
            yield idx, nextgp

    def minimal_gridded_perms(
        self, yield_non_minimal: bool = False
    ) -> Iterator[GriddedPerm]:
        """
        Yield all minimal gridded perms on the tiling.

        If `yield_non_minimal` is True, then it will yield some gridded perms
        that are non-minimal, found by the initial_gp method. Even though it
        may not be minimal, this is useful when trying to determine whether or
        not a tiling is empty.
        """
        if not self.requirements:
            if GriddedPerm.empty_perm() not in self.obstructions:
                yield GriddedPerm.empty_perm()
            return
        if len(self.requirements) == 1:
            yield from self.requirements[0]
            return

        # a priority queue sorted by length of the gridded perm. This is
        # ensured by overwriting the __lt__ operator in the Info class.
        queue: List[QueuePacket] = []
        heapify(queue)

        initial_gps_to_auto_yield: Dict[int, Set[GriddedPerm]] = defaultdict(set)
        yielded: Set[GriddedPerm] = set()
        for gp in self._prepare_queue(queue):
            if yield_non_minimal:
                yielded.add(gp)
                yield gp
            else:
                initial_gps_to_auto_yield[len(gp)].add(gp)

        def yielded_subgridded_perm(gp: GriddedPerm) -> bool:
            """Return True if a subgridded perm was yielded."""
            return gp.contains(*yielded)

        def _process_work_packet(
            qpacket: QueuePacket, queue: List[QueuePacket]
        ) -> Iterator[GriddedPerm]:
            # now we try inserting a new point into the cell
            for (cell, localised) in self._get_cells_to_try(qpacket):
                # The `localised` bool tells us if we inserted into
                # a cell as it didn't contain the patterns in the
                # cell. If not, then we update the last cell, to
                # ensure we only insert into greater or equal cells
                next_cell = qpacket.last_cell if localised else cell
                # Insert a point in every way possible into cell, to the right
                # of any previously placed points in that cell.
                for idx, nextgp in self.insert_point(
                    qpacket.gp, cell, qpacket.mindices.get(cell, 0)
                ):
                    # The work_packets_done is used to ensure the same
                    # task is not processed twice.
                    key = (nextgp, qpacket.gps, next_cell)
                    if key in work_packets_done:
                        continue
                    work_packets_done.add(key)
                    # Only work with gridded permutations avoiding the
                    # obstructions, and those which we haven't yielded a
                    # subgridded permutation.
                    if yielded_subgridded_perm(
                        nextgp
                    ) or not self.satisfies_obstructions(nextgp, must_contain=cell):
                        continue
                    # Update the nextgp about the patterns that are
                    # contained in the subgridded permutation gp.
                    self.known_patts[nextgp].update(self.known_patts[qpacket.gp])
                    # If it satisfies the requirements, then it is a
                    # a minimal gridded permutation
                    if self.satisfies_requirements(nextgp):
                        # Keep track to ensure we don't yield a gridded
                        # perm containing it.
                        yielded.add(nextgp)
                        yield nextgp
                    else:
                        # Update the minimum index that we inserted a
                        # a point into each cell.
                        next_mindices = {
                            c: i if i <= idx else i + 1
                            for c, i in qpacket.mindices.items()
                            if c != cell
                        }
                        next_mindices[cell] = idx + 1
                        # Add the work to the queue
                        heappush(
                            queue,
                            QueuePacket(
                                nextgp,
                                qpacket.gps,
                                next_cell,
                                localised,
                                next_mindices,
                            ),
                        )

        work_packets_done: Set[WorkPackets] = set()
        curr_len = -1
        while queue:
            # take the next gridded permutation of the queue, together with the
            # theoretical counts to create a gridded permutation containing
            # each of gps.
            qpacket = heappop(queue)
            # if gp was one of the initial_gps that satisfied obs/reqs, but
            # we weren't sure at the time if it was minimal, then now is the
            # time to check and yield
            if curr_len != len(qpacket.gp):
                new_len = len(qpacket.gp)
                initial_to_yield = chain.from_iterable(
                    initial_gps_to_auto_yield.pop(i, tuple())
                    for i in range(curr_len + 1, new_len + 2)
                )
                for gp in initial_to_yield:
                    if not yielded_subgridded_perm(gp):
                        yielded.add(gp)
                        yield gp
                curr_len = len(qpacket.gp)
            yield from _process_work_packet(qpacket, queue)
        # We yield any initial left after working on the work packet.
        # This happens in particular if you don't have any WP.
        initial_to_yield = chain.from_iterable(
            initial_gps_to_auto_yield.pop(i, tuple())
            for i in sorted(initial_gps_to_auto_yield)
        )
        for gp in initial_to_yield:
            if not yielded_subgridded_perm(gp):
                yielded.add(gp)
                yield gp



#!/usr/bin/env python
import os

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding="utf-8").read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name="tilings",
    version=get_version("tilings/__init__.py"),
    author="Permuta Triangle",
    author_email="permutatriangle@gmail.com",
    description="A Python library for gridded permutations and tilings.",
    license="BSD-3",
    keywords=(
        "permutation perm gridded pattern tiling avoid contain" "occurrences grid class"
    ),
    url="https://github.com/PermutaTriangle/Tilings",
    project_urls={
        "Source": "https://github.com/PermutaTriangle/Tilings",
        "Tracker": "https://github.com/PermutaTriangle/Tilings/issues",
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    long_description=read("README.rst"),
    install_requires=[
        "comb-spec-searcher==4.0.0",
        "permuta==2.1.0",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    entry_points={"console_scripts": ["tilescope=tilings.cli:main"]},
)



from collections import Counter
from itertools import chain, product
from random import randint
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple

from sympy import Eq, Expr, Function, Number, Symbol, var

from comb_spec_searcher import Constructor, Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import Rule
from comb_spec_searcher.typing import (
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling
from tilings.algorithms import FactorWithInterleaving
from tilings.assumptions import TrackingAssumption
from tilings.misc import partitions_iterator

from .factor import assumptions_to_add, interleaving_rows_and_cols

Cell = Tuple[int, int]


class AddAssumptionsConstructor(Constructor):
    """
    The constructor used to count when a new variable is added.
    """

    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        new_parameters: Iterable[str],
        extra_parameters: Dict[str, str],
    ):
        #  parent parameter -> child parameter mapping
        self.extra_parameters = extra_parameters
        #  the paramater that was added, to count we must sum over all possible values
        self.new_parameters = tuple(new_parameters)
        self._child_param_map = self._build_child_param_map(parent, child)

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        rhs_func = rhs_funcs[0]
        subs: Dict[Symbol, Expr] = {
            var(child): var(parent) for parent, child in self.extra_parameters.items()
        }
        for k in self.new_parameters:
            subs[k] = Number(1)
        return Eq(lhs_func, rhs_func.subs(subs, simultaneous=True))

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        assert len(subterms) == 1
        return self._push_add_assumption(n, subterms[0], self._child_param_map)

    @staticmethod
    def _push_add_assumption(
        n: int,
        child_terms: Callable[[int], Terms],
        child_param_map: ParametersMap,
    ) -> Terms:
        new_terms: Terms = Counter()
        for param, value in child_terms(n).items():
            new_terms[child_param_map(param)] += value
        return new_terms

    def _build_child_param_map(self, parent: Tiling, child: Tiling) -> ParametersMap:
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        child_param_to_parent_param = {v: k for k, v in self.extra_parameters.items()}
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
            tuple()
            if param in self.new_parameters
            else (parent_param_to_pos[child_param_to_parent_param[param]],)
            for param in child.extra_parameters
        )
        return self.build_param_map(
            child_pos_to_parent_pos, len(parent.extra_parameters)
        )

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        for param, gps in subobjs[0](n).items():
            yield self._child_param_map(param), (gps,)

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        subrec = subrecs[0]
        subsampler = subsamplers[0]
        random_choice = randint(1, parent_count)
        new_params = {self.extra_parameters[k]: val for k, val in parameters.items()}
        res = 0
        for values in product(*[range(n + 1) for _ in self.new_parameters]):
            for k, val in zip(self.new_parameters, values):
                new_params[k] = val
            res += subrec(n, **new_params)
            if random_choice <= res:
                return (subsampler(n, **new_params),)

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and len(other.new_parameters) == len(self.new_parameters)
            and AddAssumptionsConstructor.extra_params_equiv(
                (self.extra_parameters,), (other.extra_parameters,)
            ),
            None,
        )


class AddAssumptionsStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(self, assumptions: Iterable[TrackingAssumption], workable=False):
        self.assumptions = tuple(set(assumptions))
        super().__init__(
            ignore_parent=False,
            inferrable=True,
            possibly_empty=False,
            workable=workable,
        )

    @staticmethod
    def can_be_equivalent() -> bool:
        return False

    @staticmethod
    def is_two_way(comb_class: Tiling):
        return False

    @staticmethod
    def is_reversible(comb_class: Tiling) -> bool:
        return False

    @staticmethod
    def shifts(
        comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0,)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        if any(assumption in comb_class.assumptions for assumption in self.assumptions):
            raise StrategyDoesNotApply("The assumption is already on the tiling.")
        return (comb_class.add_assumptions(self.assumptions),)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> AddAssumptionsConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't split the tracking assumption")
        new_parameters = [
            children[0].get_assumption_parameter(ass) for ass in self.assumptions
        ]
        return AddAssumptionsConstructor(
            comb_class,
            children[0],
            new_parameters,
            self.extra_parameters(comb_class, children)[0],
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(ass)
                for ass in comb_class.assumptions
            },
        )

    def formal_step(self) -> str:
        if len(self.assumptions) == 1:
            return f"adding the assumption '{self.assumptions[0]}'"
        assumptions = ", ".join([f"'{ass}'" for ass in self.assumptions])
        return f"adding the assumptions '{assumptions}'"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        assert len(objs) == 1 and objs[0] is not None
        yield objs[0]

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        return (obj,)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d["assumptions"] = [ass.to_jsonable() for ass in self.assumptions]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AddAssumptionsStrategy":
        assumptions = [TrackingAssumption.from_dict(ass) for ass in d["assumptions"]]
        return cls(assumptions)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"

    def __repr__(self):
        return (
            self.__class__.__name__
            + f"(assumptions={repr(self.assumptions)}, workable={self.workable})"
        )


class AddAssumptionFactory(StrategyFactory[Tiling]):
    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        for assumption in comb_class.assumptions:
            without = comb_class.remove_assumption(assumption)
            strategy = AddAssumptionsStrategy((assumption,))
            yield strategy(without)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return "add assumptions"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AddAssumptionFactory":
        return cls()


class AddInterleavingAssumptionFactory(StrategyFactory[Tiling]):
    def __init__(self, unions: bool = False):
        self.unions = unions

    @staticmethod
    def strategy_from_components(
        comb_class: Tiling, components: Tuple[Tuple[Cell, ...], ...]
    ) -> Iterator[Rule]:
        """
        Yield an AddAssumption strategy for the given component if needed.
        """
        cols, rows = interleaving_rows_and_cols(components)
        assumptions = set(
            ass
            for ass in chain.from_iterable(
                assumptions_to_add(cells, cols, rows) for cells in components
            )
            if ass not in comb_class.assumptions
        )
        if assumptions:
            strategy = AddAssumptionsStrategy(assumptions, workable=True)
            yield strategy(comb_class)

    # TODO: monotone?
    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        factor_algo = FactorWithInterleaving(comb_class)
        if factor_algo.factorable():
            min_comp = tuple(tuple(part) for part in factor_algo.get_components())
            if self.unions:
                for partition in partitions_iterator(min_comp):
                    components = tuple(
                        tuple(chain.from_iterable(part)) for part in partition
                    )
                    yield from self.strategy_from_components(comb_class, components)
            yield from self.strategy_from_components(comb_class, min_comp)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return "add interleaving assumptions to factor"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["unions"] = self.unions
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AddInterleavingAssumptionFactory":
        return cls(**d)



from collections import defaultdict
from itertools import chain
from typing import DefaultDict, Iterable, Iterator, List, Optional, Set, Tuple, Union

import requests
from logzero import logger

from comb_spec_searcher import (
    CombinatorialSpecification,
    CombinatorialSpecificationSearcher,
)
from comb_spec_searcher.rule_db import RuleDBForgetStrategy
from comb_spec_searcher.rule_db.abstract import RuleDBAbstract
from comb_spec_searcher.strategies import AbstractStrategy
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.typing import CombinatorialClassType, CSSstrategy
from permuta import Basis, Perm
from tilings import GriddedPerm, Tiling
from tilings.strategies import AddAssumptionFactory, RearrangeAssumptionFactory
from tilings.strategies.assumption_insertion import AddAssumptionsStrategy
from tilings.strategies.rearrange_assumption import RearrangeAssumptionStrategy
from tilings.strategy_pack import TileScopePack

__all__ = ("TileScope", "TileScopePack", "LimitedAssumptionTileScope", "GuidedSearcher")


class TileScope(CombinatorialSpecificationSearcher):
    """
    An instance of TileScope is used to build up knowledge about tilings with
    respect to the given basis.
    """

    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        ruledb: Optional[RuleDBAbstract] = None,
        expand_verified: bool = False,
        debug: bool = False,
    ) -> None:

        """Initialise TileScope."""
        if isinstance(start_class, str):
            basis = Basis.from_string(start_class)
        elif isinstance(start_class, Tiling):
            if start_class.dimensions == (1, 1):
                basis = Basis(*[o.patt for o in start_class.obstructions])
            start_tiling = start_class
        else:
            try:
                basis = Basis(*start_class)
            except TypeError as e:
                raise ValueError(
                    "start class must be a string, an iterable of Perm or a tiling"
                ) from e

        if not isinstance(start_class, Tiling):
            start_tiling = Tiling(
                obstructions=[GriddedPerm.single_cell(patt, (0, 0)) for patt in basis]
            )

        if start_tiling.dimensions == (1, 1):
            logger.debug("Fixing basis in basis aware verification strategies.")
            strategy_pack = strategy_pack.add_basis(basis)
        strategy_pack = strategy_pack.setup_subclass_verification(start_tiling)

        super().__init__(
            start_class=start_tiling,
            strategy_pack=strategy_pack,
            ruledb=ruledb,
            expand_verified=expand_verified,
            debug=debug,
        )


class LimitedAssumptionTileScope(TileScope):
    """
    A subclass of Tilescope that allows a limit to be set on the maximum number of
    assumptions that appear on any tiling in the universe.
    """

    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        max_assumptions: int,
        **kwargs
    ) -> None:
        super().__init__(start_class, strategy_pack, **kwargs)
        self.max_assumptions = max_assumptions

    def _expand(
        self,
        comb_class: CombinatorialClassType,
        label: int,
        strategies: Tuple[CSSstrategy, ...],
        inferral: bool,
    ) -> None:
        """
        Will expand the combinatorial class with given label using the given
        strategies, but only add rules whose children all satisfy the max_assumptions
        requirement.
        """
        if inferral:
            self._inferral_expand(comb_class, label, strategies)
        else:
            for strategy_generator in strategies:
                for start_label, end_labels, rule in self._expand_class_with_strategy(
                    comb_class, strategy_generator, label
                ):
                    if all(
                        len(child.assumptions) <= self.max_assumptions
                        for child in rule.children
                    ):
                        self.add_rule(start_label, end_labels, rule)


class GuidedSearcher(TileScope):
    def __init__(
        self,
        tilings: Iterable[Tiling],
        basis: Tiling,
        pack: TileScopePack,
        *args,
        **kwargs
    ):
        self.tilings = frozenset(t.remove_assumptions() for t in tilings)
        super().__init__(basis, pack, *args, **kwargs)
        for t in self.tilings:
            class_label = self.classdb.get_label(t)
            is_empty = self.classdb.is_empty(t, class_label)
            if not is_empty:
                self.classqueue.add(class_label)

    def _expand(
        self,
        comb_class: Tiling,
        label: int,
        strategies: Tuple[CSSstrategy, ...],
        inferral: bool,
    ) -> None:
        if comb_class.remove_assumptions() not in self.tilings:
            return
        return super()._expand(comb_class, label, strategies, inferral)

    @classmethod
    def from_spec(
        cls, specification: CombinatorialSpecification, pack: TileScopePack
    ) -> "GuidedSearcher":
        tilings = specification.comb_classes()
        root = specification.root
        return cls(tilings, root, pack)

    @classmethod
    def from_uri(cls, URI: str) -> "GuidedSearcher":
        response = requests.get(URI)
        spec = CombinatorialSpecification.from_dict(response.json()["specification"])
        pack = TileScopePack.from_dict(response.json()["pack"]).make_tracked()
        return cls.from_spec(spec, pack)


class TrackedSearcher(LimitedAssumptionTileScope):
    """
    A TileScope that only adds underlying tilings to the queue, but expands all
    assumption tilings with the strategies that apply to the underlying tiling
    immediately.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tilings_from_underlying: DefaultDict[int, Set[int]] = defaultdict(set)
        self.tracking_strategies = [
            AddAssumptionFactory(),
            RearrangeAssumptionFactory(),
        ]
        self.tracked_expanded: Set[int] = set()
        self.retroactively_expanded: Set[int] = set()
        # TODO: keep self._strats on the ruledb, and avoid storing strats twice.
        self._strats: DefaultDict[int, List[AbstractStrategy]] = defaultdict(list)

    def store_strategy(self, label: int, strategy: AbstractStrategy) -> None:
        self._strats[label].append(strategy)

    def get_old_strategies(self, label: int) -> Tuple[CSSstrategy, ...]:
        return tuple(self._strats[label])

    def add_rule(
        self, start_label: int, end_labels: Tuple[int, ...], rule: AbstractRule
    ) -> None:
        """
        Add the rule to the searcher

        - try to verify children combinatorial classes
        - set workability of combinatorial classes
        - symmetry expand combinatorial classes
        - add underlying class to classqueue, and do the expansion for any assumption
        tilings with same underlying tiling
        """
        for comb_class, child_label in zip(rule.children, end_labels):
            underlying_label = (
                self.classdb.get_label(comb_class.remove_assumptions())
                if comb_class.assumptions
                else child_label
            )
            if underlying_label != child_label:
                self.tilings_from_underlying[underlying_label].add(child_label)
            if self.symmetries and child_label not in self.symmetry_expanded:
                self._symmetry_expand(comb_class, child_label)
            if rule.workable:
                # add the underlying label to the queue
                self.classqueue.add(underlying_label)
            if (
                underlying_label != child_label
                and child_label not in self.retroactively_expanded
            ):
                self.retroactively_expanded.add(child_label)
                # apply all rules in ruledb to child_label
                old_strategies = self.get_old_strategies(underlying_label)
                self._expand(comb_class, child_label, old_strategies, False)
            # apply tracking strategies
            if comb_class.assumptions and child_label not in self.tracked_expanded:
                self.tracked_expanded.add(child_label)
                self._expand(comb_class, child_label, self.tracking_strategies, False)
            if not rule.inferrable:
                self.classqueue.set_not_inferrable(underlying_label)
            if not rule.possibly_empty:
                # this is shortcutting some empty checks by ruledb,
                # so should be done to both underlying and child labels
                self.classdb.set_empty(child_label, empty=False)
                if underlying_label != child_label:
                    self.classdb.set_empty(underlying_label, empty=False)
            underlying_tiling = (
                comb_class.remove_assumptions()
                if comb_class.assumptions
                else comb_class
            )
            # calls add rule recursively, so will add verification
            # rules for all with underlying tiling
            self.try_verify(underlying_tiling, underlying_label)
        if rule.ignore_parent:
            self.classqueue.set_stop_yielding(start_label)

        # update all with same underlying label
        self.ruledb.add(start_label, end_labels, rule)
        if not isinstance(
            rule.strategy,
            (RearrangeAssumptionStrategy, AddAssumptionsStrategy),
        ):
            self.store_strategy(start_label, rule.strategy)
            for label in list(self.tilings_from_underlying[start_label]):
                assumption_tiling = self.classdb.get_class(label)
                self._expand(assumption_tiling, label, (rule.strategy,), False)


class ForgetTrackedSearcher(TrackedSearcher):
    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        **kwargs
    ):
        self.strategies: List[CSSstrategy] = list(
            chain(
                strategy_pack.ver_strats,
                strategy_pack.initial_strats,
                strategy_pack.inferral_strats,
                *strategy_pack.expansion_strats,
            )
        )
        self._strat_indices: DefaultDict[int, int] = defaultdict(int)
        kwargs["ruledb"] = kwargs.get("ruledb", RuleDBForgetStrategy())
        super().__init__(start_class, strategy_pack, **kwargs)

    def store_strategy(self, label: int, strategy: AbstractStrategy) -> None:
        """We do nothing as instead we track in the _expand method."""

    def get_old_strategies(self, label: int) -> Tuple[CSSstrategy, ...]:
        return tuple(
            self.strategies[idx]
            for idx, bit in enumerate(bin(self._strat_indices[label])[-1:1:-1])
            if bit == "1"
        )

    def _expand_class_with_strategy(
        self,
        comb_class: CombinatorialClassType,
        strategy_generator: CSSstrategy,
        label: Optional[int] = None,
        initial: bool = False,
    ) -> Iterator[Tuple[int, Tuple[int, ...], AbstractRule]]:
        if not isinstance(
            strategy_generator, (AddAssumptionFactory, RearrangeAssumptionFactory)
        ):
            try:
                idx = self.strategies.index(strategy_generator)
                assert isinstance(label, int)
                self._strat_indices[label] |= 1 << idx

            except ValueError:
                pass
        yield from super()._expand_class_with_strategy(
            comb_class, strategy_generator, label, initial
        )
